#include "adapt.h"
#include "gpg.h"
#include "convert.h"


//-------------------------------------------------------------------------------------------------
//write_problem
//descrizione degli oggetti del dominio
//PARAMETRI
//ProbFile	Puntatore al file del dominio
//object_list	Lista degli oggetti e loro tipo
//inertia_list	Lista fatti non modificabili
//facts		Fatti del problema corrente
//goals		Lista obiettivi del problema
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem (FILE *ProbFile, fact_list objects_list, token_list inertia_list, token_list facts, token_list goals)
{
//Scrive sul file del problema l'intestazione e gli oggetti del problema
    write_problem_header (ProbFile);

//Scrive gli oggetti definiti nel problema
    write_problem_objects (ProbFile, objects_list);

//Scrive intestazione fatti iniziali
    fprintf (ProbFile,"\t(:init\n");

//Scrive i fatti non modificabili da alcun operatore
    write_problem_inertias (ProbFile, inertia_list);
    
//Scrive sul file del problema la descrizione dell'istante iniziale ( un insieme di relazioni)
    write_problem_initial_facts (ProbFile,facts);

//Scrive sul file del problema la descrizione del goal ( anch'esso un insieme di relazioni)
    write_problem_goal_facts (ProbFile,goals);

//Definizione del problema conclusa, chiude il file
    fclose (ProbFile);

    return 1;
}



//-------------------------------------------------------------------------------------------------
//write_problem_header
//descrizione degli oggetti specifici del problema
//PARAMETRI
//ProbFile	Puntatore al file del problema
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem_header (FILE* ProbFile)
{
//NOTA: La funzione dovr essere modificata per trattare domini diversi da lofistics
    fprintf (ProbFile,"(define (problem\tmine)\n");
    fprintf (ProbFile,"\t(:domain\tthis-domain)\n");
    return 1;
}

//-------------------------------------------------------------------------------------------------
//write_problem_objects
//descrizione degli oggetti specifici del problema
//PARAMETRI
//ProbFile	Puntatore al file del problema
//objects_list	Puntatore alla lista degli oggetti
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem_objects (FILE* ProbFile, fact_list objects_list)
{
    fact_list i_fct;
    token_list i_tok, first_obj;
    char *typename;
//intestazione dichiarazione oggetti
    fprintf (ProbFile,"\t(:objects\n");

//scorre i tipi degli oggetti: per ogni item della fact_list contiene in testa il nome del tipo e quindi gli oggetti di quel tipo
    for (i_fct = objects_list; i_fct; i_fct = i_fct->next)
    {
	fprintf(ProbFile,"\t\t");
//tiene via la prima stringa (il tipo), che in pddl va messa per ultima
	typename = i_fct->item->item;
//TOGLIERE
	if (strcmp (typename,"object") ==0)
	    typename[3]='a';
//FINE TOGLIERE

//puntatore al primo oggetto di quel tipo
	first_obj = i_fct->item->next;
//scorre e stampa gli oggetti di quel tipo
	for (i_tok = first_obj; i_tok; i_tok = i_tok->next)
	    fprintf (ProbFile,"%s ",i_tok->item);
//stampa il tipo degli oggetti
	fprintf(ProbFile,"- %s\n",typename);
    }
//chiude parentesi dichiarazione oggetti
    fprintf (ProbFile,"\t)\n");
    return 1;
}


//-------------------------------------------------------------------------------------------------
//write_problem_inertias
//descrizione delle relazioni non modificabili da alcun operatore
//PARAMETRI
//ProbFile	Puntatore al file del problema
//inertia_list	Puntatore alla lista degli inertia
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem_inertias (FILE* ProbFile, token_list inertia_list)
{
    token_list i_tok;
    char begstr[90];
    int i;
    

//stampa la lista dei fatti inerziali
    for (i_tok=inertia_list; i_tok; i_tok=i_tok->next)
    {
	strcpy (begstr, i_tok->item);
//Sostituisce '_' con uno spazio	
	for (i = 0; begstr[i] != '\0'; i++)
	    if (begstr[i] == '_')
		begstr[i] = ' ';
	fprintf (ProbFile,"\t\t(%s)\n", begstr);
    }

    return 1;
}


//-------------------------------------------------------------------------------------------------
//write_problem_initial_facts
//descrizione dei fatti iniziali del problema
//PARAMETRI
//ProbFile	Puntatore al file del problema
//list		lista dei fatti iniziali
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem_initial_facts (FILE* ProbFile, token_list list)
{
    token_list i_token;
    char begstr[100];
    int i;


//Per ogni oggetto della lista iniziale, aggiungi una relazione
    for ( i_token = list; i_token; i_token = i_token->next )
    {
	fprintf (ProbFile,"\t\t(");

//Fa una copia dell'oggetto della lista
	strcpy(begstr,i_token->item);

//Sostituisce '_' con uno spazio	
	for (i = 0; begstr[i] != '\0'; i++)
	    if (begstr[i] == '_')
		begstr[i] = ' ';

//Stampa su file la stringa 'pulita'	
	fprintf(ProbFile, begstr);

	fprintf (ProbFile,")\n");
    }

//Chiusura parentesi fatti iniziali
    fprintf (ProbFile,"\t)\n");

    return 1;
}


//-------------------------------------------------------------------------------------------------
//write_problem_goal_facts
//descrizione degli obiettivi del problema
//PARAMETRI
//ProbFile	Puntatore al file del problema
//list		Lista dei goal da raggiungere
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem_goal_facts (FILE* ProbFile, token_list list)
{
    token_list i_token;
    char begstr[100];
    int i;

//Apertura contesto goals
    fprintf (ProbFile,"\t(:goal\t(and\n");

//Per ogni oggetto della lista...
    for ( i_token = list; i_token; i_token = i_token->next )
    {
//Apre la parentesi per l'oggetto corrente
	fprintf (ProbFile,"\t\t\t(");

//Fa una copia dell'oggetto della lista
	strcpy(begstr,i_token->item);

//Sostituisce '_' con uno spazio	
	for (i = 0; begstr[i] != '\0'; i++)
	    if (begstr[i] == '_')
		begstr[i] = ' ';

//Stampa su file la stringa 'pulita'	
	fprintf(ProbFile, begstr);

//Chiude la parentesi
	fprintf (ProbFile,")\n");
    }
//Fine file del problema
    fprintf (ProbFile,"\t\t)\n\t)\n)\n");

    return 1;
}

