/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */

/* 	$Id: eq_preproc.c,v 1.3 1997/10/02 07:25:40 ipp Exp $	 */

#ifndef lint
static char vcid[] = "$Id: eq_preproc.c,v 1.3 1997/10/02 07:25:40 ipp Exp $";
#endif /* lint */


#include"ipp.h"/* defines, data structures, fn prototypes, global variables */

/* main function for reducing given list to
 * instantiable operators;
 * called by main ( in file plan.c )
 */
void remove_uninstantiable_ops( op_list *ops, fact_list constants )

{

  if ( *ops ) {/* still something to do */
    if ( remove_able( *ops, constants ) ){
      /* remove this operator by assigning it the value of
       * it's successor
       */
      *ops = (*ops)->next;
      /* remove uninstantiable operators in rest of list */
      remove_uninstantiable_ops( ops, constants );
    } else {
      /* leave this op untouched and continue with rest of list */
      remove_uninstantiable_ops( &((*ops)->next), constants );
    }
  }

}

/* helpfunction for remove_uninstantiable_ops:
 *   returns TRUE gdw op cannot be instantiated
 */
BOOLEAN remove_able( op_list op, fact_list constants )

{

  fact_list parameter;/* index for currently examined parameter */
  BOOLEAN result = FALSE;/* that's the result of the test */

  /* check the parameterlist of op */
  for( parameter = op->params; parameter; parameter = parameter->next ) {
    if ( cannot_instantiate( parameter->item, constants ) ) {
      result = TRUE;/* found an uninstantiable parameter */
      break;
    }
  }

  /* remove the uninstantiable effects; if these are all effects,
   * op can as well be removed
   */
  remove_effects( &(op->effects), constants );
  if ( !op->effects ) result = TRUE;

  return result;

}

void remove_effects( effect_list *effects, fact_list constants )

{

  fact_list parameter;

  if ( *effects ) {/* still something to do */
    for ( parameter = (*effects)->quantified_variables;
          parameter; parameter = parameter->next )
      if ( cannot_instantiate( parameter->item, constants ) ) break;
    if ( parameter ) {
      /* remove this effect by assigning it the value of
       * it's successor
       */
      *effects = (*effects)->next;
      /* remove uninstantiable operators in rest of list */
      remove_effects( effects, constants );
    } else {
      /* leave this effect untouched and continue with rest of list */
      remove_effects( &((*effects)->next), constants );
    }
  }

}



/* helpfunction for remove_able:
 *   returns TRUE gdw there's no matching constant for parameter
 */
BOOLEAN cannot_instantiate( token_list param, fact_list constants )

{

  fact_list i_constant;/* index for currently examined constant */
  BOOLEAN result = TRUE;/* that's the result of the test */


  /* check all the constants in given list */
  for( i_constant = constants; i_constant; i_constant = i_constant->next ) {
    if ( strcmp( i_constant->item->next->item, param->next->item ) == SAME ) {
      result = FALSE;/* found one that matches */
      break;
    }
  }

  return result;
}



/* some helpfunctions - allocate memory for new list elements */

meta_fact_list new_meta_fact_list()
{
  meta_fact_list result;
  result = ( meta_fact_list ) malloc( sizeof( meta_fact_list_elt ) );
  result->item = NULL; result->next = NULL;
  return result;
}

/* returns an integer describing if the predicate
   is 'eq', 'not-eq' or something else */
int check_eq( token f )
{
  if ( strncmp( f, EQ_STR, strlen( EQ_STR ) ) == SAME )
      return EQ;
  if ( strncmp( f, NOT_EQ_STR, strlen( NOT_EQ_STR ) ) == SAME )
      return NOT_EQ;
  return ANY_PRED;
}


/* checks if a token consists of two equal strings
   dividided by an underscore '_' */
BOOLEAN args_eq( token two_args )
{
  int pos;
  
  for( pos=1; two_args[pos-1]!='_'; pos++ )
    ;
  /* step through both parts of the string until
     end is reached (=equal) or difference
     is detected (=different) */
  while ( two_args[pos] == two_args[0] )
    two_args++;
  if ( two_args[pos] == 0 )
    return TRUE;
  else 
    return FALSE;
}



/* tests if conds are contained in the initial facts 
   or if an 'eq()' or 'not-eq()' predicate is false */
BOOLEAN cond_in_initial( token_list conds, token_list initials )
{
  token_list i_c, i_i;
  int pred_type;

  for ( i_c = conds; i_c; i_c=i_c->next ) 
    {
      /* check if eq or not-eq */
      pred_type = check_eq( i_c->item );
      if ( pred_type == EQ )
	{ /* predicate is 'equal' but arguments are different */
	  if ( !args_eq( (i_c->item)+strlen( EQ_STR ) ) )
	    return FALSE;
	}
      else if ( pred_type == NOT_EQ )
	{ /* predicate is 'not-equal' but arguments are the same */
	  if ( args_eq( (i_c->item)+strlen( NOT_EQ_STR ) ) )
	    return FALSE;
	}
      else
	{ /* predicate is neither 'eq' nor 'not-eq',
	   so check if in initial state */
	  for ( i_i = initials; i_i; i_i=i_i->next )
	    if ( strcmp( i_i->item, i_c->item ) == SAME ) break;
	  if ( !i_i ) return FALSE;
	}
    }
  return TRUE;
}



/* make copy of a token_list and save the end of that
   list (...to attach further elements) */
token_list copy_tl_return_end( token_list tl, token_list *end )
{
  token_list t, start;
  
  start = t = new_token_list();
  for( ; tl; tl = tl->next )
    {
      t->next = new_token_list();
      t = t->next;
      t->item = tl->item;
      if ( !tl->next )
	{
	  *end = t;
	}
    }
  return start->next;

}



/* build a fact_list in which every token_list describes one
   possibility to fulfill the goal description. This is done
   by combining all possibilities for a certain 
   goal decription factor_mfl->item with all yet found 
   incomplete instatiated goals - in the each fac_list represents
   one goal description (conjunctive goals), each of these
   fact_lists is a set of token_lists being the possible
   instatiations for that goal (disjunctive goals), and
   each token_list consists of several tokens, which must
   be fulfilled with the same parameter instatiation 
   (conjunctive goals again) */
fact_list rec_build_exgoal_fl( fact_list product_fl, 
			       meta_fact_list factor_mfl )
{
  /* product_fl is a list of incomplete goal descriptions, 
     which is splitted and multiplied by the yet missing set of
     goal descriptions */
  fact_list o;
  token_list end = NULL;
  fact_list result, act;

  if ( !factor_mfl )
    return product_fl;
  /* split and copy fact_lists in product_mfl */
  result = act = new_fact_list();
  for ( ; product_fl; product_fl=product_fl->next )
    {
      for ( o=factor_mfl->item; o; o=o->next )
	{
	  act->next = new_fact_list();
	  if ( !(product_fl->item) )
	    { /* no goal facts yet - usually this means
		 that there are only existentially
		 quantified goals */
	      act->next->item = o->item;
	    }
	  else
	    { /* if other goals exist find end of list
		 and add new goals */
	      act->next->item = copy_tl_return_end( product_fl->item, &end );
	      end->next  = o->item;
	    }
	  act = act->next;
	}
    } 
  return rec_build_exgoal_fl( result->next, factor_mfl->next );
}



/* remove quantifiers by instatiating goals with all possible
   constants. If existentially quantified goals exist that
   offer different possibilities to be fulfilled, build
   'goal operators' to reach the final goal in a number of
   different ways.
   Another purpose of that functions is to eliminate eq() and
   not-eq() facts in goal conditions - this is done
   (invisible in this function) during the evaluation of
   cond_in_initials(). Remark: same thing is done in
   instantiate.c when checking operator preconditions */
token_list dequantify_goals( effect_list raw_goal_facts, op_list *ops )
{
  inst_effect_list inst_goal_list;
  token_list result = NULL;
  token_list act_goal;
  int act_quant;
  effect_list next_raw_goal;
  int ex_goal_no = FALSE;
  int act_op_no = 1;
  op_list act_op;
  fact_list ex_goals, act_ex_goal;
  meta_fact_list ex_goal_list, act_ex_goal_list;
  token pEXGOAL_REACHED;
  char EXGOAL_REACHED[] = "ex-goal-reached";
  char EXGOAL_OPNAME[] = "ex";

  pEXGOAL_REACHED = ( token ) calloc( strlen( EXGOAL_REACHED )+1, sizeof( char ) );
  strcpy( pEXGOAL_REACHED, EXGOAL_REACHED );
  result = act_goal = new_token_list();
  ex_goal_list = act_ex_goal_list = new_meta_fact_list();
  while ( raw_goal_facts )
    { /* the first fact contains only one token specifying the quantifier... */
      act_quant = raw_goal_facts->quantified_variables->item->item[0];
      raw_goal_facts->quantified_variables =  
	raw_goal_facts->quantified_variables->next;
      /* only process one goal at time*/
      next_raw_goal = raw_goal_facts->next; 
      raw_goal_facts->next = NULL;

      if ( act_quant == QUANTIFIERS[ ALL_QUANT ]  )
	{ /* universally quantified goal */
	  inst_goal_list = delete_ALL_quantifying( raw_goal_facts, 
						   orig_constant_list, NULL );
	  while ( inst_goal_list )
	    {
	      /* if unconditional or conditions in initial state 
		 just add goal to list */
	      if ( NULL == inst_goal_list->conditions ||
		   cond_in_initial( inst_goal_list->conditions, initial_facts ) )
		{
		  act_goal->next = inst_goal_list->add_effects;
		  while ( act_goal->next )
		    act_goal = act_goal->next;
		}
	      inst_goal_list = inst_goal_list->next;
	    }
	}
      else if ( act_quant == QUANTIFIERS[ EX_QUANT ] )
	{ /* existentially quantified goal */
	  inst_goal_list = delete_ALL_quantifying( raw_goal_facts, 
						   orig_constant_list, NULL );
	  ex_goals = act_ex_goal = new_fact_list();
	  /* search for goal conditions contained in initial state */
	  for (; inst_goal_list; inst_goal_list = inst_goal_list->next )
	    {
	      /* if unconditional or conditions in initial state 
		 just add goal to list */
	      if ( NULL == inst_goal_list->conditions ||
		   cond_in_initial( inst_goal_list->conditions, initial_facts ) )
		{ 
		  act_ex_goal->next = new_fact_list();
		  act_ex_goal->next->item = inst_goal_list->add_effects;
		  act_ex_goal = act_ex_goal->next;
		}
	    }
	  /* ex_goals is a fact_list in which every token_list contains
	     instantiated predicates which are conjuctive goals, only one
	     of these conjunctive goals must be reached, i.e. only one token_list
	     contains facts which are really to be made true -
	     important: list begins at ex_goals->next */
	  if ( ex_goals->next ) 
	    { /* at least one goal possibility found */ 
	      if ( !( ex_goals->next->next ) )
		{ /* only one possibility to fulfill this goal, so we can
		     include the conjuctive goals into the usual list 
		     of goals */
		  act_goal->next = ex_goals->next->item;
		  for( ; act_goal->next; act_goal = act_goal->next )
		    ;
		}
	      else
		{ /* there are at least two goals for which the precond 
		     is fulfilled => so we must build ex-goal operators - but 
		     for now only a set of fact_lists is needed: beginning
		     at ex_goal_list->next, the actual position being 
		     act_ex_goal_list, it holds a fact_list for every
		     given existentially quantified goal. Each token_list
		     in a fact_list describes possible instantiations
		     for the abstract goal description */
		  ex_goal_no = TRUE;
		  act_ex_goal_list->next = new_meta_fact_list();
		  act_ex_goal_list = act_ex_goal_list->next;
		  act_ex_goal_list->item = ex_goals->next;
		  /* act_ex_goal_list->item is now the fact_list holding
		     all possibilities to reach one EX goal */
		}
	    } /* endif ( ex_goals->next ) */ 
	} /* endif EX_QUANT */
      else
	{ /* not quantified goal */
	  if ( cond_in_initial( token_list_from_fact_list( raw_goal_facts->conditions ), 
				initial_facts ) )
	    {
	      act_goal->next = 
		token_list_from_fact_list( raw_goal_facts->add_effects );
	      while ( act_goal->next )
		act_goal = act_goal->next;
	    }
	}
      raw_goal_facts = next_raw_goal;
    } /* endwhile */
  /* now we stepped through all goal descriptions and must
     decide if we need goal operators or may leave the
     list of goals (beginning at result->next) untouched */
  if ( ex_goal_no )
    { /* existentially quantified goals exist, so we must
	 build ex-goal operators, each of which possibly
	 may reach the 'goal state'. preconditions of these
	 operators describe possible combinations of ex-goals
	 and universally or not quantified goals. So
	 what we're doing here is in fact set multiplication
	 to get all possible sets of goals... */
      token_list the_only_goal;
      fact_list ex_goal_fl;
      fact_list act_pre;
      /* only one goal is needed which may be reached
	 by different goal operators */
      the_only_goal = new_token_list();
      the_only_goal->item = pEXGOAL_REACHED;
      /* we have to combine the universally or not quantified goals
	 with one possibility for reaching every ex_goal. In the
	 end we need a token_list for each possible goal state */
      ex_goal_fl = new_fact_list();
      /* begin with universally or not quantified goals */
      ex_goal_fl->item = result->next;
      /* find all combinations by recursive set multiplying -
	 same method as in rec_inst_pred() */
      ex_goal_fl = rec_build_exgoal_fl( ex_goal_fl, ex_goal_list->next );
      /* now every token_list in exgoal_fl describes one possible
	 goal state - we only have to build a new ex-goal
	 operator with this token_list as preconds */
      for( act_pre=ex_goal_fl; act_pre; act_pre=act_pre->next )
	{
	  act_op = ( op_list ) malloc( sizeof( op_list_elt ) );
	  act_op->name = ( char * ) calloc( 1+strlen(EXGOAL_OPNAME)+6, sizeof(char) );
	  sprintf( act_op->name, "%s%d", EXGOAL_OPNAME, act_op_no ); 
	  act_op->params = NULL;
	  factlist_from_tokenlist( &(act_op->preconds), act_pre->item );
	  act_op->effects = new_effect_list();
	  act_op->effects->add_effects = new_fact_list();
	  act_op->effects->add_effects->item = the_only_goal;
	  act_op->params_objects = NULL;
	  /* include new operator into operator list */
	  act_op->next = *ops;
	  *ops = act_op;
	  act_op_no++;
	}
      return the_only_goal;
    }
  else
    return result->next;
}
