/*
 *  Header GPG : estratto da ipp.h versione newgam (strips-like domains) 
 */ 

#ifndef __GPG_H
#define __GPG_H

/* hash values for gam */

#define G_HSIZE 512
#define G_HASH 511
#define INTSIZE 32
#define GPG_MAX_PLAN 1
#define NO_EXAMINE 2



#define MEMO( type ) {bytes_memo += sizeof( type );}
#define GRAPHO( type ) {bytes_o += sizeof( type );}
#define GRAPHF( type ) {bytes_f += sizeof( type );}
#define GRAPHE( type ) {bytes_e += sizeof( type ); numbereff++;}
#define GRAPHK( type ) {bytes_k += sizeof( type ); numberkan++;}
#define EXCL( type ) {bytes_excl += sizeof( type );}

#define NO_INCONS -1
#define INIT_ACT   "initial_action"
#define GOAL_ACT   "goal_action"
#define ALL_NAMES   ""
#define ALL_LEVELS  -10
#define TEMP_DELETED -1
#define FIRST_1 0x80000000           

//miamod
//Possibili limitazioni imposte al planner esterno
#define CPU_TIME 1
#define LEVELS_NUM 2
#define ACTIONS_NUM 3

//#define NORMAL_FIXPOINT
#define RELAXED_FIXPOINT
//miamod
struct GPG_COMMAND
{
    int limitation_type;
} gpg_command;

//finemiamod

/* simply a goal
 */
typedef struct GOAL 
{

  token name;
  token_list objects;

  struct GOAL *next;
} *goal_list, goal_list_elt;


/* this structure contains the agenda: a list of goal sets;
 * st and hits are used for ordering sets( in compute.c )
 */
typedef struct GOALSET
{

  goal_list set;

  int st;
  int hits;

  BOOLEAN intermediates;

  struct GOALSET *left;
  struct GOALSET *right;

} *goal_set_list, goal_set_list_elt;


/* the following make up the graph structure for ordering goals:
 * we have a list op_nodes that contains all actions and a
 * hashtable goal_table that contains all the goals that are
 * added by them. the whole stuff is connected by pointers so that
 * every needed information is accessible easily.
 */
typedef struct OPEDGE *o_edge_list, o_edge_list_elt;
typedef struct EFEDGE *e_edge_list, e_edge_list_elt;
typedef struct GOEDGE *g_edge_list, g_edge_list_elt;
typedef struct EFNODE *e_node_list, e_node_list_elt;

typedef struct OPEXCL *o_excl_list, o_excl_list_elt;
typedef struct GOEXCL *g_excl_list, g_excl_list_elt;


/* an action;
 */
typedef struct OPNODE 
{

  token name;/* as usual */
  g_edge_list preconds;/* the preconds; ( pointers to goals in the fact table ) */
  e_node_list effects;/* the effects */
#ifdef NORMAL_FIXPOINT
  o_excl_list excl_at[MAX_PLAN];
#endif
#ifdef RELAXED_FIXPOINT
  o_excl_list excl_at[GPG_MAX_PLAN];
#endif

  int uid_block;
  unsigned int uid_mask;

  BOOLEAN is_noop;
  BOOLEAN is_used;
#ifdef RELAXED_FIXPOINT
  BOOLEAN is_examined;
#endif
  struct OPNODE *next;
} *o_node_list, o_node_list_elt;


/* an effect;
 * in contradiction to the graphplan graph for planning,
 * in GAM we have separated effect nodes;
 * seems to be more natural for ipp style operators
 */
struct EFNODE 
{

  BOOLEAN on;/* says if this effect can be used currently; see compute.c */
  o_node_list op;/* the op it belongs to */

  g_edge_list conditions;/* should be self explanatory */
  g_edge_list add_edges;
  g_edge_list del_edges;

  /* these for fixpoint */
  BOOLEAN is_new;
  token_list del_names;

  struct EFNODE *next;
};



// types for causal links

typedef struct ACTION_AT_TIME
{

  token action;
  int time;

} action_at_time;


typedef struct ACTION_AT_TIME_LIST
{

  action_at_time of;
  struct ACTION_AT_TIME_LIST *next;

} *action_at_time_list, action_at_time_list_elt;



typedef struct CAUSAL_LINK
{

  action_at_time from;
  action_at_time to;
  token fact;

  struct CAUSAL_LINK *previous;	
  struct CAUSAL_LINK *next;

} *causal_link_list, causal_link_list_elt;


// struct added for planning with gpg

typedef struct GPG_STRUCT
{

  BOOLEAN is_true;
  BOOLEAN is_precond;
  BOOLEAN is_added;
  BOOLEAN is_deleted;
  short int is_goal;

  action_at_time_list precond;
  action_at_time_list add_effect;
  action_at_time_list del_effect;

} gpg_struct;


// ...and the goals

typedef struct GONODE 
{

  token name;

  /* technical stuff for compute.c */

#ifdef NORMAL_FIXPOINT
  g_excl_list excl_at[MAX_PLAN];
#endif
#ifdef RELAXED_FIXPOINT
  g_excl_list excl_at[GPG_MAX_PLAN];
#endif

  g_edge_list false_facts;
  g_edge_list true_facts;

  int Tcount;
  int Fcount;
  BOOLEAN is_true;
  
  int st;
  int hits;

  int uid_block;
  unsigned int uid_mask;

   /* these are for intermediates */
  BOOLEAN is_initial;
  BOOLEAN old_selected;
  BOOLEAN new_selected;
  int count;

  /* these for gam_plan */
  int initial_count;
  BOOLEAN curr_true;
  BOOLEAN made_true;
  BOOLEAN made_false;

  /* and these are the actions and effects that have to do with this goal */
  o_edge_list preconds;
  e_edge_list effconds;
  e_edge_list add_edges;
  e_edge_list del_edges;

  /* these for fixpoint */
  BOOLEAN has_noop;
  BOOLEAN is_new;

  gpg_struct gpg;

  struct GONODE *next;
} *g_node_list, g_node_list_elt;

#ifdef RELAXED_FIXPOINT
g_node_list gpg_fact_list[MAX_MAX_NODES];
#endif

/* edge lists for everyone...
 */

struct OPEDGE 
{

  o_node_list endpt;

  struct OPEDGE *next;
};

struct EFEDGE 
{

  e_node_list endpt;

  struct EFEDGE *next;
};

struct GOEDGE 
{

  g_node_list endpt;

  struct GOEDGE *next;
};

struct OPEXCL 
{

  char *name;

  int exclusive_vect[NUMINTS];
};

struct GOEXCL 
{

  char *name;

  int exclusive_vect[NUMINTS];
};


/* this makes up the hashtable for GAM goals
 */
typedef g_node_list g_table_type[G_HSIZE];
typedef o_node_list o_table_type[G_HSIZE];

/* just for improving readability of some algorithms
 */
typedef g_node_list g_array_type[MAX_GOALS];

/* to order sets, they are stored in an set array
 */
typedef goal_set_list goal_set_list_array_type[MAX_GOALS];

/* these are technical for intermediates,
 * i.e. to see if there has been a circle: compare bit strings
 */

typedef int bit_string[NUMINTS];

typedef struct BITSTRINGLIST
{

  bit_string string;

  struct BITSTRINGLIST *prev;
  struct BITSTRINGLIST *next;
} *bit_string_list, bit_string_list_elt;


/*
 * function prototypes
 */


/* in file gam.c */

void make_gam_goal_facts( goal_set_list list, BOOLEAN first_try );
void setup_gam_initial_facts( int time );
void evaluate( int time );
void permut_evaluate( int time );
void permutate( int time, int n );
void simple_evaluate( int time );
void op_evaluate( vertex_list op );
void linearize_subplan( goal_set_list curr );
int fac( int N );

token_list gam_token_list_from_token( token str );
fact_list fact_list_from_token_list( token_list t );
void initialise_set_list( token_list goal_facts );
void print_gam_set_list( void );
BOOLEAN oversize( goal_list goals );
void insert( goal_set_list lp, goal_set_list rp, goal_list goals );
int gam_hash( char *key );
g_node_list insert_go( token t );
g_node_list lookup_go( token t );
o_node_list insert_op( token t );
o_node_list lookup_op( token t );
BOOLEAN applicable( token_list list );
g_edge_list insert_g_edge( g_node_list g, g_edge_list list );
o_edge_list insert_o_edge( o_node_list o, o_edge_list list );
e_edge_list insert_e_edge( e_node_list e, e_edge_list list );
void print_ef_node( e_node_list e );
void print_go_op_nodes( void );
BOOLEAN contain( g_edge_list l1, g_edge_list l2 );
void gam_quicksort( g_array_type a, int l, int r );
void gam_set_quicksort( goal_set_list_array_type a, int l, int r );
void set_gam_go_uid( g_node_list g );
void set_gam_op_uid( o_node_list o );
void print_goal_set( goal_set_list l );
void print_goal_list( goal_list l );
BOOLEAN gam_are_there_non_exclusive( token_list facts, int time );
BOOLEAN gam_get_them_non_exclusive( token_list facts, int time,
				    g_array_type *ft, int *num );
e_node_list effect_is_already_there( o_node_list o, inst_effect_list e,
				     int time );
BOOLEAN gam_go_are_mutex( g_node_list g1, g_node_list g2, int time );
#ifdef NORMAL_FIXPOINT
void gam_go_make_exclusive( g_node_list g1, g_node_list g2, int time );
#endif
#ifdef RELAXED_FIXPOINT
int gam_go_make_exclusive( g_node_list g1, g_node_list g2, int time );
#endif

BOOLEAN gam_op_are_mutex( o_node_list o1, o_node_list o2, int time );
void gam_op_make_exclusive( o_node_list o1, o_node_list o2, int time );
BOOLEAN no_prev_time( g_node_list g, int time );
void print_F( g_node_list g );
void print_T( g_node_list g );

void compute_agenda( goal_set_list lp, goal_set_list rp );
BOOLEAN set_reachable( goal_set_list l1, goal_set_list l2 );
void setup_FT( g_node_list g1 );
BOOLEAN reachable( g_node_list g );
BOOLEAN is_added( g_node_list g );
void setup_O( g_node_list g1 );
void unsetup_O( g_node_list g1 );

goal_list intermediate( goal_list goals );
void generate_intersections( void );
BOOLEAN circle_occured( bit_string_list curr );
goal_list make_goal_list( bit_string vec );


/* in file gpg.c */

void print_g_table();
void print_o_table();
void initialize_gpg_fixpoint(void);
int forward_propagation(int time_init, int time_end);
void compute_goals(int goal_time, int time_end);
void compute_CL_goals(int goal_time);
void compute_causal_links();
void plan_check();


/* in file fixpoint.c */

int build_fixpoint( void );
void gam_apply_operator( operator_list o, int time );
void gam_find_mutex_ops_and_insert_deletes( int time );
int gam_find_mutex_facts( int time );
BOOLEAN gam_facts_are_exclusive( g_node_list g1, g_node_list g2, int time );
void insert_noops( void );


/*
 * global variables
 */


/* globally important variables */

/* domain constants */

token_list incr_goal_facts;/* used for gam */


/* the actual graph: two arrays of hashtables
 * in the current implemetation, we stote the graphs for each subproblem,
 * that's why the graph is now a two dimensional array.
 * gam_active_part contains the number of the sub problem 
 * that is currently solved.
 */

int gam_active_part;
int max_small_set;/* limit for splitting up goal sets */
BOOLEAN do_inters;/* compute intermediates or not ? */
BOOLEAN linearize;/* linearize sub plans ? */
BOOLEAN do_true_set;


/* option preset values */

int debug;/* level of GAM debugging info */
BOOLEAN do_true_set;

/* gam main stuff */
goal_set_list start, end;/* start and end of goal set list */

/* compute stuff */
o_table_type o_table;
g_table_type g_table;

int gam_total_num_ops;
int gam_total_num_goals;

BOOLEAN gam_fixpoint_reached;
int gam_fixpoint_level;

long bytes_o;
long bytes_f;
long bytes_e;
long bytes_k;
long bytes_excl;
long numbereff;
long numberkan;
long numberexclkan;
long bytes_graph;
long bytes_memo;

/* intermediate stuff */
bit_string_list first_string, last_string;
bit_string temp_string;


// causal links

causal_link_list causal_links;

#endif __GPG_H
